
#pragma once
#include <cmath>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/constraints/common/constraintsparameters.hh>

#include <dune/phasefield/initial_fn.hh>

#define GRID_TOP_THINSTRIP 1


template<typename DomainSize>
class LeftBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  LeftBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
    if ( xg[0] > ds.domainEps)
    {
      return false;
    }
    return true;
  }
};

template<typename DomainSize>
class BottomBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  BottomBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
      if ( xg[1]>ds.bottom + ds.domainEps)
      {
        return false;
      }
      return true;
  }
};

template<typename DomainSize>
class LeftRightBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  LeftRightBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    Dune::FieldVector<typename I::ctype, I::coorddimension>
      xg = intersection.geometry().global( coord );
      if (xg[1]>ds.bottom + ds.domainEps && xg[1]<ds.top - ds.domainEps)
      {
        return true;
      }
      return false;
  }
};

template<typename DomainSize>
class AllBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  AllBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    return true;
  }
};

template<typename DomainSize>
class NoBoundary : public Dune::PDELab::DirichletConstraintsParameters
{
private:
  const DomainSize ds;
public:
  NoBoundary(const DomainSize d) : ds(d)
  {
  }

  template<typename I>
  bool isDirichlet(const I & intersection, const Dune::FieldVector<typename I::ctype, I::mydimension> & coord) const
  {
    return false;
  }
};

template<typename GV, typename RF, typename Params, int dim>
class VelocityInitial_InOutFlow :
  public Dune::PDELab::AnalyticGridFunctionBase<
  Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim>,
  VelocityInitial_InOutFlow<GV,RF,Params,dim> >
{
  const Params_Time& time;
  RF v;
  RF yS;


public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,dim> Traits;

  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::RangeType RangeType;

  VelocityInitial_InOutFlow(const GV & gv, const Params & params) :
    Dune::PDELab::AnalyticGridFunctionBase<Traits,VelocityInitial_InOutFlow<GV,RF,Params,dim> > (gv)
    , time(params.time)
  {
    //TODO: Use right scaling (now small coordinates)
    v = params.initial.ptree.get("pConst",(RF)0.01)/(params.YScaling*params.YScaling) /2 /*Domain Width*/ /params.phys.mu;
  }

  template <typename T>
  void setTime(T t){
    time = t;
  }

  inline void evaluateGlobal(const DomainType & x, RangeType & y) const
  {
    y[0] = std::max(0.0, v/2.0*(x[1]-0.3)*((2*GRID_TOP_THINSTRIP-0.3)-x[1]));
  }
};

template<typename GV, typename RF, typename Params>
class PInitial_InOutFlow :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PInitial_InOutFlow<GV,RF,Params> >
{
  RF pConst;
public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PInitial_InOutFlow(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PInitial_InOutFlow<GV,RF,Params> > (gv)
     {
       pConst = params.initial.ptree.get("pConst",(RF)0.01);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    y = 0;
    /*if(x[0]<1e-7)
    {
      y=pConst/(yS*yS);
    }*/
  }
};

template<typename GV, typename RF, typename Params, const int Phase>
class PhiInitial_Thinstrip_Layers :
  public Dune::PDELab::AnalyticGridFunctionBase<
    Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1>, PhiInitial_Thinstrip_Layers<GV,RF,Params,Phase> >
{
  RF eps;
  RF dSolid;
  RF d1;
  const Params & pref;

public:
  typedef Dune::PDELab::AnalyticGridFunctionTraits<GV,RF,1> Traits;

  PhiInitial_Thinstrip_Layers(const GV & gv, const Params & params) :
     Dune::PDELab::AnalyticGridFunctionBase<Traits,PhiInitial_Thinstrip_Layers<GV,RF,Params,Phase> > (gv),
     pref(params)
     {
       eps = params.eps;
       dSolid = params.initial.ptree.get("dSolid",0.3);
       d1 = params.initial.ptree.get("d1",0.3);
     }

  inline void evaluateGlobal(const  typename Traits::DomainType & x, typename Traits::RangeType & y) const
  {
    RF rf = pref.tw.dw.eval_shape(1/eps*(x[1] - dSolid));
    RF rf2 = pref.tw.dw.eval_shape(1/eps*(x[1] - dSolid - d1));
    if(Phase == 1)
    {
      y = rf2;
      return;
    }
    y=rf*(1-rf2);
  }

  void ApplyReactionDt(const Params & params)
  {
    RF uIn = params.initial.ptree.get("uInflow",(RF)0.5);
    dSolid += params.reaction.eval(uIn)*params.time.dt;
    d1 -= params.reaction.eval(uIn)*params.time.dt;
    std::cout << "New dSolid " << dSolid << " , changed by " << params.reaction.eval(uIn) << " times dt";
  }
};
